//This work is licensed under a Creative Commons Attribution 3.0 Unported License.
//Original version by Daniel Davis, 2013
//danieldavis.com
//This chapter is based on work by Tiemen Rapati:
//http://www.tiemenrapati.com/blog/?p=439

import geomerative.*;
import java.util.LinkedList;

class ch0 extends ch_abstract
{
  public void ch0()
  {
  }

  ////////////////////////////////////////////////////////
  //  Main functions
  ////////////////////////////////////////////////////////

  private float segmentlengte = 1;
  private int advance = 2;//6


  private float slowopacity = 50.0F;
  private float fastopacity = 90.0F;
  private float factor = 1.0F;

  private RPoint[][] text_points;
  private Pica[][] picas;

  public void alphabet(String fileName)
  {
    setup();
    beginPDFRecord(fileName);

    strokeCap(SQUARE);
    strokeWeight(1);
    noFill();
    
    for(int i = 0; i < words.length; i++)
    {
      pushMatrix();
      translate( width/2, step_size * (i+1));
      makeWord(words[i], 400);
      popMatrix();
    }
    endPDFRecord();
  }

  private void makeWord(String word, int iterations)
  {
    this.picas = new Pica[0][0];
    setTargets(word, font);

    for (int n = 0; n < iterations; n++)//400
    {
      updateTargets();

      for (int i = 0; i < this.picas.length; i++) 
      {
        for (int j = 0; j < this.picas[i].length; j++) 
        {
          this.picas[i][j].update();
        }
      }
    }

    draw();
  }

  private void draw()
  {

    //New Drawing method that links the points up. 
    for (int i = 0; i < this.picas.length; i++) 
    {
      LinkedList<Pica>[] pts = (LinkedList<Pica>[]) new LinkedList[20];
      for (int j = 0; j < 20; j++) pts[j] = new LinkedList<Pica>();

      for (int j = 0; j < this.picas[i].length; j++) 
      {
        //every 20th pica belongs to a curve, so we send the 20th picas to their own array. 
        pts[j%20].add(picas[i][j]);
      }

      for (LinkedList<Pica> curvePts : pts)
      {
        if (curvePts.size() < 4) continue;

        //draw the segments of the curves individually so they can change color
        for (int j = 0; j < curvePts.size(); j++)
        {
          int[] sel = {
            j-3, j-2, j-1, j
          };

          //use this to stop the array getting back on itself. 
          for (int k = 0; k < sel.length; k++)
          {
            if (sel[k] < 0) sel[k] = curvePts.size() + sel[k];
          }

          float angle = atan((curvePts.get(sel[1]).getLoc().x - curvePts.get(sel[2]).getLoc().x) / (curvePts.get(sel[1]).getLoc().y - curvePts.get(sel[2]).getLoc().y));

          float angleNormal = (abs(angle)) / (1.570796F);
          color c = colorMix(angleNormal);

          float f2 = map(curvePts.get(sel[1]).getVel().mag(), 0.0F, 20.0F, this.slowopacity * this.factor, this.fastopacity * this.factor);

          stroke(c, f2);

          beginShape();
          //parent.stroke(parent.colorMix(parent.random(1)));

          for (int k = 0; k < sel.length; k++)
          {
            curveVertex(curvePts.get(sel[k]).getLoc().x, curvePts.get(sel[k]).getLoc().y);
          }
          endShape();
        }
      }
    }
  }


  private void setTargets(String str, RFont font)
  {
    //RCommand.setSegmentLength(this.segmentlengte);
    //RCommand.setSegmentator(1);
    //this.text_points = geometry.text_to_points(str, font); 
    this.text_points = returnPoints(str, font);

    this.picas = new Pica[text_points.length][];

    for (int i = 0; i < picas.length; i++)
    {
      picas[i] = new Pica[this.text_points[i].length];

      for (int j = 0; j < this.text_points[i].length; j++)
      {
        float f1 = j % 20;
        float f2 = 0.5F + j % 10 / 20.0F;

        Pica localPica = new Pica(j, 0.0F, 0.0F, f1, f2, 10);
        localPica.setTarget(new PVector(this.text_points[i][j].x, this.text_points[i][j].y));
        this.picas[i][j] = localPica;
      }
    }
  }

  private void updateTargets()
  {
    for (int i = 0; i < this.picas.length; i++)
    {
      for (int j = 0; j < this.picas[i].length; j++)
      {
        this.picas[i][j].setTargetIndex(this.picas[i][j].getTargetIndex() + this.advance);
        if (this.picas[i][j].getTargetIndex() >= this.text_points[i].length) 
        {
          this.picas[i][j].setTargetIndex(this.picas[i][j].getTargetIndex() % this.text_points[i].length);
        }
        else if (this.picas[i][j].getTargetIndex() < 0) 
        {
          this.picas[i][j].setTargetIndex(this.picas[i][j].getTargetIndex() + this.text_points[i].length);
        }

        PVector localPVector = new PVector(this.text_points[i][this.picas[i][j].getTargetIndex()].x, this.text_points[i][this.picas[i][j].getTargetIndex()].y);
        this.picas[i][j].setTarget(localPVector);
      }
    }
  }


  private RPoint[][] returnPoints(String word, RFont font)
  {
    RGroup localRGroup = font.toGroup(word);
    RCommand.setSegmentLength(this.segmentlengte);
    RCommand.setSegmentator(RCommand.UNIFORMLENGTH);
    RPoint[][] arrayOfRPoint = localRGroup.getPointsInPaths();

    return arrayOfRPoint;
  }
}


class Pica
{
  private PVector loc;
  private PVector[] history;
  private PVector vel;
  private PVector acc;

  private PVector target;
  private int target_index;

  private float maxforce;
  private float maxspeed;


  Pica(int _target_index, float _x, float _y, float _maxspeed, float _maxforce, int historyLength)
  {
    this.target_index = _target_index;

    this.acc = new PVector(0.0F, 0.0F);
    this.vel = new PVector(0.0F, 0.0F);
    this.loc = new PVector(_x, _y);

    this.maxspeed = _maxspeed;
    this.maxforce = _maxforce;
    this.history = new PVector[historyLength];
  }

  public void setTarget(PVector t)
  {
    target = t;
  }

  public int getTargetIndex()
  {
    return target_index;
  }

  public void setTargetIndex(int ti)
  {
    target_index = ti;
  }

  public PVector getLoc()
  {
    return loc;
  }

  public PVector getVel()
  {
    return vel;
  }



  public void update()
  {
    this.acc.add(steer());

    this.vel.add(this.acc);

    this.vel.limit(this.maxspeed);
    this.loc.add(this.vel);

    this.acc.mult(0.0F);

    for (int i = this.history.length - 1; i > 0; i--) 
    {
      this.history[i] = this.history[(i - 1)];
    }
    this.history[0] = new PVector(this.loc.x, this.loc.y);
  }

  private PVector steer()
  {
    PVector localPVector2 = PVector.sub(this.target, this.loc);
    float f = localPVector2.mag();
    PVector localPVector1;
    if (f > 0.0F)
    {
      localPVector2.normalize();

      if (f < 100.0F) {
        localPVector2.mult(this.maxspeed * (f / 100.0F));
      }
      else {
        localPVector2.mult(this.maxspeed);
      }

      localPVector1 = PVector.sub(localPVector2, this.vel);
      localPVector1.limit(this.maxforce);
    }
    else
    {
      localPVector1 = new PVector(0.0F, 0.0F);
    }

    return localPVector1;
  }
}



/*
  
 @Override
 public void alphabet(String fileName)
 {
 setup();
 
 pushMatrix();
 beginPDFRecord(fileName);
 
 translate( width/2, step_size);
 RGroup grp = font.toGroup( "ABCDE" );
 skew_font(grp, 0.05f);
 
 translate( 0, step_size);
 grp = font.toGroup( "FGHIJK" );
 skew_font(grp, 0.02f);
 
 translate( 0, step_size);
 grp = font.toGroup( "LMNOP" );
 skew_font(grp, 0.01f);
 
 translate( 0, step_size);
 grp = font.toGroup( "QRSTU" );
 skew_font(grp, -0.03f);
 
 translate( 0, step_size);
 grp = font.toGroup( "VWXYZ" );
 skew_font(grp, -0.2f);
 
 endPDFRecord();
 
 popMatrix();
 
 }
 
 private void skew_font(RGroup grp, float factor)
 {
 RGeomElem[] letters = grp.elements;
 
 int no_of_steps = 20;
 
 //shear the letters into maximum position
 for (int n = 0; n < letters.length; n++)
 {      
 RGeomElem letter = letters[n];
 letter.shear((no_of_steps + 1) * factor * -1, 0);
 }
 
 //start with most sheared and then step down into at front letter (f==0)
 for (float f = no_of_steps; f >= 0; f--)
 {
 float fNormal = f * 1f/no_of_steps;
 
 color c = colorMix(1-fNormal);
 noFill();
 stroke(c, 100 - (fNormal * 100));
 strokeWeight(1);
 
 //if we are at the last step, draw top letters in blue
 if (f == 0)
 {
 fill(blue, 180);
 stroke(c, 255);
 }
 
 //loop through letters, draw them and make them less sheared.
 for (int n = 0; n < letters.length; n++) 
 {        
 RShape shp = new RShape();
 
 RGeomElem letter = letters[n];
 
 letter.shear(factor, 0);
 letter.draw();
 }
 }
 }
 }*/
